/*
 *  Simple trackball-like motion ripped off from projtex.c (written by
 *  David Yu and David Blythe).  See the GLUT sample programs (advanced).
 *
 *  Copyright (C) 1997 Nate Robins (ndr@pobox.com)
 *
 *  email: ndr@pobox.com
 *  www: http://www.pobox.com/~ndr
 *
 *  This program is freely distributable without licensing fees and is
 *  provided without guarantee or warrantee expressed or implied. This
 *  program is not in the public domain.
 *
 *
 *  Usage:
 *  
 *  o  call tbInit() in before any other tb call
 *  o  call tbReshape() from the reshape callback
 *  o  call tbMatrix() to get the trackball matrix rotation
 *  o  call tbStartMotion() to begin trackball movememt
 *  o  call tbStopMotion() to stop trackball movememt
 *  o  call tbMotion() from the motion callback
 *
 *  Typical setup:
 *
 *
    void
    init(void)
    {
      tbInit(GLUT_MIDDLE_BUTTON);
      . . .
    }

    void
    reshape(int width, int height)
    {
      tbReshape(width, height);
      . . .
    }

    void
    display(void)
    {
      glPushMatrix();

      tbMatrix();
      . . . draw the scene . . .

      glPopMatrix();
    }

    void
    mouse(int button, int state, int x, int y)
    {
      tbMouse(button, state, x, y);
      . . .
    }

    void
    motion(int x, int y)
    {
      tbMotion(x, y);
      . . .
    }

    int
    main(int argc, char** argv)
    {
      . . .
      init();
      glutReshapeFunc(reshape);
      glutDisplayFunc(display);
      glutMouseFunc(mouse);
      glutMotionFunc(motion);
      . . .
    }
 *
 *
 */
/* 
Ok.. I have made trackball class and added translate and scale routines to the code.
If you improve it , please contribute..
Levan shoshiashvili; levan@lae.ictsu.tsu.edu.ge
*/

#ifndef _TRACKBALL_H
#define _TRACKBALL_H
#include <GL/glut.h>
const double   PI = 3.14159265358979323846;
const double D2RAD = 0.01745329251994329576;
#define sqr(a)	 ((a)*(a))
class Trackball
{
public:
	Trackball();
	~Trackball();
void tbInit(GLuint button);
GLfloat   tb_translate[3];
void tbReshape(int width, int height);
GLfloat   tb_initial[4][4];
void tbMouse(int button, int state, int x, int y);
void tbMatrix();
void tbTranslate();
void tbMotion(int x, int y);	GLfloat   tb_angle;

private:

void _tbPointToVector(int x, int y, int width, int height, float v[3]);
static void _tbAnimate(void);
void _tbStartMotion(int x, int y, int button, int time);
void _tbStopMotion(int button, unsigned time);



enum _tb_transform_mode{tb_rotate,tb_scale,tb_shift} tb_transform_mode;
GLuint    tb_lasttime;
GLfloat   tb_lastposition[3];

GLfloat   tb_axis[3];
GLfloat   tb_transform[4][4];
GLuint    tb_width;
GLuint    tb_height;

GLint     tb_button;
GLint     tb_shift_modifier;
GLint     tb_scale_modifier;
GLboolean tb_tracking;
GLboolean tb_animate ;
GLuint tb_last_x;
GLuint tb_last_y;

};
#endif