#include "Trackball.h"
#include <math.h>
#include <assert.h>
#include "Matrix4.hpp"
Trackball::Trackball()
{
tb_lasttime=0;
tb_lastposition[3];
tb_angle = 0.0;
tb_axis[3];
tb_transform[4][4];
tb_width=300;
tb_height=300;
tb_translate[3];
tb_button = -1;
tb_shift_modifier = -1;
tb_scale_modifier = -1;
tb_tracking = GL_FALSE;
tb_animate = GL_TRUE;

}
Trackball::~Trackball()
{

}
 void Trackball::_tbPointToVector(int x, int y, int width, int height, float v[3])
{
  float d, a;

  /* project x, y onto a hemi-sphere centered within width, height. */
  v[0] = (2.0 * x - width) / width;
  v[1] = (height - 2.0 * y) / height;
  d = sqrt(v[0] * v[0] + v[1] * v[1]);
  v[2] = cos((3.14159265 / 2.0) * ((d < 1.0) ? d : 1.0));
  a = 1.0 / sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
  v[0] *= a;
  v[1] *= a;
  v[2] *= a;
}

 void Trackball::_tbAnimate(void)
{
  glutPostRedisplay();
}

void Trackball::_tbStartMotion(int x, int y, int button, int time)
{
  assert(tb_button != -1);
  tb_tracking = GL_TRUE;
  tb_lasttime = time;
  _tbPointToVector(x, y, tb_width, tb_height, tb_lastposition);
}

void Trackball::_tbStopMotion(int button, unsigned time)
{
  assert(tb_button != -1);
  tb_tracking = GL_FALSE;

  if (time == tb_lasttime && tb_animate) {
    glutIdleFunc(_tbAnimate);
  } else {
    tb_angle = 0.0;
    if (tb_animate)
      glutIdleFunc(0);
  }
}

void Trackball::tbInit(GLuint button)
{
  tb_button =GLUT_LEFT_BUTTON;
  tb_angle = 0.0;

  /* put the identity in the trackball transform */

  glPushMatrix();
  glLoadIdentity();
  glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *)tb_transform);
  tb_translate[0] = 0;
  tb_translate[1] = 0;
  tb_translate[2] = 0;
  glPopMatrix();
}

void Trackball::tbMatrix()
{
  assert(tb_button != -1);
switch(tb_transform_mode)
  {
case tb_shift:
	{
//glMultMatrixf((GLfloat *)tb_transform);

 glPushMatrix();
  glLoadIdentity();

 glTranslated(tb_translate[0], tb_translate[1],0);
    glMultMatrixf((GLfloat*)tb_transform);
   glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *)tb_transform);
   glPopMatrix();
  glMultMatrixf((GLfloat *)tb_transform);

	}
  break;

 case tb_scale:
	 glPushMatrix();
	 glLoadIdentity();
	 float scale;;

		scale = pow(1.1,tb_translate[1]);

		glMultMatrixf((GLfloat*)tb_transform);
	glScalef(scale, scale, scale);

	 glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat*)tb_transform);
	 glPopMatrix();
	 glMultMatrixf((GLfloat*)tb_transform);


	break;
case tb_rotate:
//	glLoadIdentity();

glPushMatrix();
  glLoadIdentity();

glMultMatrixf((GLfloat *)tb_transform);

{

  glRotatef(tb_angle, tb_axis[0], tb_axis[1], tb_axis[2]);
}


  glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *)tb_transform);
  glPopMatrix();
 glMultMatrixf((GLfloat *)tb_transform);
break;
default:
	break;
//tb_scale=GL_FALSE;
  }

}

void Trackball::tbReshape(int width, int height)
{
  assert(tb_button != -1);

  tb_width  = width;
  tb_height = height;

}

void Trackball::tbMouse(int button, int state, int x, int y)
{
  assert(tb_button != -1);

  if (state == GLUT_DOWN && button == GLUT_LEFT_BUTTON)
  {
	  tb_transform_mode=tb_rotate;
	  _tbStartMotion(x, y, button, glutGet(GLUT_ELAPSED_TIME));
  }


  if (state == GLUT_DOWN && button == GLUT_LEFT_BUTTON && GLUT_ACTIVE_SHIFT==glutGetModifiers())
  {
	  tb_transform_mode=tb_shift;

  }
  if (state == GLUT_DOWN && button == GLUT_LEFT_BUTTON && GLUT_ACTIVE_CTRL==glutGetModifiers())
  {
	  tb_transform_mode=tb_scale;

  }
 if (state == GLUT_UP && button == GLUT_LEFT_BUTTON)
 {
	   tb_transform_mode=tb_rotate;
	 _tbStopMotion(button, glutGet(GLUT_ELAPSED_TIME));
 }
}

void Trackball::tbMotion(int x, int y)
{
  GLfloat current_position[3], dx, dy, dz;

  assert(tb_button != -1);

  if (tb_tracking == GL_FALSE)
    return;

  _tbPointToVector(x, y, tb_width, tb_height, current_position);

  /* calculate the angle to rotate by (directly proportional to the
     length of the mouse movement */
  dx = current_position[0] - tb_lastposition[0];
  dy = current_position[1] - tb_lastposition[1];
  dz = current_position[2] - tb_lastposition[2];
  tb_angle = 90.0 * sqrt(dx * dx + dy * dy + dz * dz);


if (tb_scale)
{
	tb_translate[2] = dy;
}
if (tb_shift)
{
	tb_translate[0] = dx*tb_width/10;
	tb_translate[1] = dy*tb_height/10;
	tb_translate[2] = 0;
}
if (tb_rotate)
{
	tb_translate[0] = 0;
	tb_translate[1] = 0;
	tb_translate[2] = 0;
}



  /* calculate the axis of rotation (cross product) */
  tb_axis[0] = tb_lastposition[1] * current_position[2] -
               tb_lastposition[2] * current_position[1];
  tb_axis[1] = tb_lastposition[2] * current_position[0] -
               tb_lastposition[0] * current_position[2];
  tb_axis[2] = tb_lastposition[0] * current_position[1] -
               tb_lastposition[1] * current_position[0];

  /* reset for next time */
  tb_lasttime = glutGet(GLUT_ELAPSED_TIME);
  tb_lastposition[0] = current_position[0];
  tb_lastposition[1] = current_position[1];
  tb_lastposition[2] = current_position[2];

  /* remember to draw new position */
  glutPostRedisplay();
}
