//***************************************************************************//
//                                                                           //
//           name :      4x4 Matrix Class Templaet:				             //
//           source:     Matrix2.hpp                                         //
//           version:    1.0   (Aug 4, 2003)                                 //
//           use:        Defines 4x4 Matrix class                            //                                                                  //
//           developers:  Alexander Y. Bijamov                               //
//                        Levan Shoshiashvili                                //
//                                                                           //
//  Copyright (c) 2003, Laboratory of Applied Electrodynamics (LAE)          //
//                      Tbilisi State University                             //
//                                                                           //
//***************************************************************************//
//                                NOTE:                                      //
// 1.division by 0 in inv() is not checked.so check det() first              //
// 2. to make identity matrix do diag(1) ;)                                  //                                                                          //
// 3.you need vec3^mat4x4 template? this is in another story(hpp file) ;)    //
// 4. WARNING: there is no bound checking for .set functions.                //
//                                                                           //
//***************************************************************************//
#ifndef LAE_44_MATRIX
#define LAE_44_MATRIX
#include <math.h>
#include <assert.h>
#pragma warning (disable : 4514)
template <class T>

class mat4x4
{
public:
	T m00,m01,m02,m03;
	T m10,m11,m12,m13;
	T m20,m21,m22,m23;
	T m30,m31,m32,m33;

	mat4x4(const T _m00=T(), const T _m01=T(), const T _m02=T(),const T _m03=T(),
		   const T _m10=T(), const T _m11=T(), const T _m12=T(),const T _m13=T(),
		   const T _m20=T(), const T _m21=T(), const T _m22=T(),const T _m23=T(),
		   const T _m30=T(), const T _m31=T(), const T _m32=T(),const T _m33=T()):
			m00(_m00),m01(_m01),m02(_m02),m03(_m03),
			m10(_m10),m11(_m11),m12(_m12),m13(_m13),
			m20(_m20),m21(_m21),m22(_m22),m23(_m23),
			m30(_m30),m31(_m31),m32(_m32),m33(_m33){};



	template <class R>
		mat4x4(const mat4x4<R>& mat):
	m00(T(mat.m00)),m01(T(mat.m01)),m02(T(mat.m02)),m03(T(mat.m03)),
    m10(T(mat.m10)),m11(T(mat.m11)),m12(T(mat.m12)),m13(T(mat.m13)),
	m20(T(mat.m20)),m21(T(mat.m21)),m22(T(mat.m22)),m23(T(mat.m23)),
	m30(T(mat.m30)),m31(T(mat.m31)),m32(T(mat.m32)),m33(T(mat.m33))
	{
		#pragma message ("Type Conversion in mat4x4<T>")
	};

	template <class R>
	  	operator mat4x4<R>()
	{
		#pragma message ("Type Conversion in mat4x4<T>")
		return mat4x4<R>(R(m00),R(m01),R(m02),R(m03),
			             R(m10),R(m11),R(m12),R(m13),
						 R(m20),R(m21),R(m22),R(m23),
						 R(m30),R(m31),R(m32),R(m33)
						 );
	};

	mat4x4<T>& operator=(const mat4x4<T>& mat)
	{
		if (&mat == this)
			return *this;
		m00 = mat.m00;m01 = mat.m01;m02 = mat.m02;m03 = mat.m03;
		m10 = mat.m10;m11 = mat.m11;m12 = mat.m12;m13 = mat.m13;
		m20 = mat.m20;m21 = mat.m21;m22 = mat.m22;m23 = mat.m23;
		m30 = mat.m30;m31 = mat.m31;m32 = mat.m32;m33 = mat.m33;
		return *this;
	}
	mat4x4<T>& operator+=(const mat4x4<T>&mat)
	{
		m00 += mat.m00;m01 += mat.m01;m02 += mat.m02;m03 += mat.m03;
		m10 += mat.m10;m11 += mat.m11;m12 += mat.m12;m13 += mat.m13;
		m20 += mat.m20;m21 += mat.m21;m22 += mat.m22;m23 += mat.m23;
		m30 += mat.m30;m31 += mat.m31;m32 += mat.m32;m33 += mat.m33;
		return *this;
	}
	mat4x4<T>& operator-=(const mat4x4<T>&mat)
	{
		m00 -= mat.m00;m01 -= mat.m01;m02 -= mat.m02;m03 -= mat.m03;
		m10 -= mat.m10;m11 -= mat.m11;m12 -= mat.m12;m13 -= mat.m13;
		m20 -= mat.m20;m21 -= mat.m21;m22 -= mat.m22;m23 -= mat.m23;
		m30 -= mat.m30;m31 -= mat.m31;m32 -= mat.m32;m33 -= mat.m33;
		return *this;
	}

	//template <class S>
  	mat4x4<T>& operator*=(const T& a)
	{
		m00 *= a;m01 *= a;m02 *= a;m03 *= a;
		m10 *= a;m11 *= a;m12 *= a;m13 *= a;
		m20 *= a;m21 *= a;m22 *= a;m23 *= a;
		m30 *= a;m31 *= a;m32 *= a;m33 *= a;
		return *this;
	}
	//template <class S>
	 mat4x4<T>& operator/=(const T& a)
	{
		m00 /= a;m01 /= a;m02 /= a;m03 /= a;
		m10 /= a;m11 /= a;m12 /= a;m13 /= a;
		m20 /= a;m21 /= a;m22 /= a;m23 /= a;
		m30 /= a;m31 /= a;m32 /= a;m33 /= a;
		return *this;
	}

	mat4x4<T>  operator-() const
	{
		return mat4x4<T>(-m00,-m01,-m02,-m03,
			             -m10,-m11,-m12,-m13,
						 -m20,-m21,-m22,-m23,
						 -m30,-m31,-m32,-m33
						 );
	};
	mat4x4<T>  operator+() const {return *this;};
	T det()
	{
		return (
			m00*(m11*m22*m33+m31*m12*m23+m21*m32*m13-m31*m22*m13-m21*m12*m33-m32*m23*m11)
           -m10*(m01*m22*m33+m31*m02*m23+m21*m32*m03-m31*m22*m03-m21*m02*m33-m32*m23*m01)
		   +m20*(m01*m12*m33+m11*m32*m03+m02*m13*m31-m31*m12*m03-m11*m02*m33-m32*m13*m01)
		   -m30*(m01*m12*m23+m02*m13*m21+m11*m22*m03-m21*m12*m03-m11*m02*m23-m22*m13*m01));
	};
	void diag(T d1,T d2,T d3, T d4)
	{
		m00=d1,m01=0,m02=0,m03=0,
		m10=0,m11=d2,m12=0,m13=0,
		m20=0,m21=0,m22=d3,m23=0;
		m20=0,m21=0,m22=0,m23=d4;
	}
	T trase()
	{
		return (m00+m11+m22+m33);
	}
	template <class R>
	void set( R const  *_m)
	{

			m00=_m[0];m01=_m[4];m02=_m[8];m03=_m[12];
			m10=_m[1];m11=_m[5];m12=_m[9];m13=_m[13];
			m20=_m[2];m21=_m[6];m22=_m[10];m23=_m[14];
			m30=_m[3];m31=_m[7];m32=_m[11];m33=_m[15];
	}
	T   * get()
	{
		T  _m[16];
		_m[0]=m00;_m[4]=m01;_m[8]=m02;  _m[12]=m03;
		_m[1]=m00;_m[5]=m01;_m[9]=m02;  _m[13]=m03;
		_m[2]=m00;_m[6]=m01;_m[10]=m02; _m[14]=m03;
		_m[3]=m00;_m[7]=m01;_m[11]=m02; _m[15]=m03;
		return _m;

	}
	template <class R>
	void set_row(const int i,R const  _m00,R const  _m01,R const _m02,R const _m03)
	{
		switch(i)
		{
			case 0:
				m00=(T) _m00;m01=(T) _m01; m02=(T) _m02;m03=(T) _m03;break;
			case 1:
				m10=(T) _m00;m11=(T) _m01; m12=(T) _m02;m13=(T) _m03; break;
			case 2:
				m20=(T) _m00;m21=(T) _m01; m22=(T) _m02;m23=(T) _m03; break;
			case 3:
				m30=(T) _m00;m31=(T) _m01; m32=(T) _m02;m33=(T) _m03; break;

			default: break;
	 }
	}

	T* get_row(const int i)
	{
		switch(i)
		{
			case 0:
				T m[4];
				m[0]=m00;
				m[1]=m01;
				m[2]=m02;
				m[3]=m03;
				return m; break;
			case 1:
				T m2[4];
				m2[0]=m10;
				m2[1]=m11;
				m2[2]=m12;
				m2[3]=m13;
				return m2; break;
			case 2:
				T m3[4];
				m3[0]=m20;
				m3[1]=m21;
				m3[2]=m22;
				m3[2]=m23;
				return m3; break;
			case 4:
				T m4[4];
				m4[0]=m30;
				m4[1]=m31;
				m4[2]=m32;
				m4[2]=m33;
				return m4; break;
			default: break;
		}
	}

	T* get_col(const int i)
	{
		switch(i)
		{
			case 0:
				T m[4];
				m[0]=m00;
				m[1]=m10;
				m[2]=m20;
				m[3]=m30;
				return m; break;
			case 1:
				T m2[4];
				m2[0]=m01;
				m2[1]=m11;
				m2[2]=m21;
				m2[3]=m31;
				return m2; break;
			case 2:
				T m3[4];
				m3[0]=m02;
				m3[1]=m12;
				m3[2]=m22;
				m3[3]=m32;
				return m3; break;
			case 3:
				T m4[4];
				m4[0]=m03;
				m4[1]=m13;
				m4[2]=m23;
				m4[3]=m33;
				return m4; break;
			default: break;
		}
	}
	template <class R>
void set(int i, int j,const R _val)
	{
		assert(i>-1 && i<4);
		assert(j>-1 && j<4);
		switch (i)
		{
		case 0:
			switch (j)
			{
			case 0:
				  m00=_val; break;
            case 1:
				 m01=_val; break;
            case 2:
				 m02=_val; break;
			case 3:
				 m03=_val; break;

            default: break;
			};
		case 1:
			switch (j)
			{
			case 0:
				 m10=_val; break;
            case 1:
				 m11=_val; break;
            case 2:
				 m12=_val; break;
			case 3:
				 m13=_val; break;
            default: break;
			};
		case 2:
			switch (j)
			{
			case 0:
				 m20=_val; break;
            case 1:
				 m21=_val; break;
            case 2:
				 m22=_val; break;
			case 3:
				 m23=_val; break;
            default: break;
			};
		case 3:
			switch (j)
			{
			case 0:
				 m30=_val; break;
            case 1:
				 m31=_val; break;
            case 2:
				 m32=_val; break;
			case 3:
				 m33=_val; break;
            default: break;
			};
		default:break;
		};
	}

	void set_col(const int i,const T _m00,const T _m01,const T _m02,const T _m03)
	{
		switch(i)
		{
			case 0:
				m00=(T) _m00;m10=(T) _m01;m20=(T) _m02;m30=(T) _m03; break;
			case 1:
				m01=(T) _m00;m11=(T) _m01;m21=(T) _m02;m31=(T) _m03; break;
			case 2:
				m02=(T) _m00;m12=(T) _m01;m22=(T) _m02;m32=(T) _m03; break;
			case 3:
				m03=(T) _m00;m12=(T) _m03;m22=(T) _m03;m33=(T) _m03; break;

			default: break;

	 }
	}

	T operator ()(int i, int j)
	{
		switch (i)
		{
		case 0:
			switch (j)
			{
			case 0:
				return m00; break;
            case 1:
				return m01; break;
            case 2:
				return m02; break;
			case 3:
				return m03; break;

            default: break;
			};
		case 1:
			switch (j)
			{
			case 0:
				return m10; break;
            case 1:
				return m11; break;
            case 2:
				return m12; break;
			case 3:
				return m13; break;
            default: break;
			};
		case 2:
			switch (j)
			{
			case 0:
				return m20; break;
            case 1:
				return m21; break;
            case 2:
				return m22; break;
			case 3:
				return m23; break;
            default: break;
			};
		case 3:
			switch (j)
			{
			case 0:
				return m30; break;
            case 1:
				return m31; break;
            case 2:
				return m32; break;
			case 3:
				return m33; break;
            default: break;
			};
		default:break;
		};
		return 0;
	}


	mat4x4<T> inv()
	 {

		mat4x4<T> r(*this);//first build adjoint matrix;
		r=r.adj();
		T det= m00*r.m00 + m01*r.m01 + m02*r.m02+m03*r.m03;
		r=r.transpose();
		r/=det;
		return r;
	 }
mat4x4 <T> transpose()
{
mat4x4<T> r(*this);
r.m01=m10;r.m02=m20;r.m03=m30;
r.m10=m01;r.m12=m21;r.m13=m31;
r.m20=m02;r.m21=m12;r.m23=m32;
r.m30=m03;r.m31=m13;r.m32=m23;
return r;
}
	mat4x4<T> adj()
	 {

		mat4x4<T> r;
		        r.m00=m11*m22*m33+m31*m12*m23+m21*m32*m13-m31*m22*m13-m21*m12*m33-m32*m23*m11;
				r.m01=-(m10*m22*m33+m30*m12*m23+m20*m32*m13-m30*m22*m13-m20*m12*m33-m32*m23*m10);
				r.m02=m10*m21*m33+m30*m11*m23+m20*m31*m13-m30*m21*m13-m20*m11*m33-m31*m23*m10;
				r.m03=-(m10*m21*m32+m30*m11*m22+m20*m31*m12-m30*m21*m12-m20*m11*m32-m31*m22*m10);

				r.m10=-(m01*m22*m33+m31*m02*m23+m21*m32*m03-m31*m22*m03-m21*m02*m33-m32*m23*m01);
				r.m11=(m00*m22*m33+m30*m02*m23+m20*m32*m03-m30*m22*m03-m20*m02*m33-m32*m23*m00);
				r.m12=-(m00*m21*m33+m30*m01*m23+m20*m31*m03-m30*m21*m03-m20*m01*m33-m31*m23*m00);
				r.m13=(m00*m21*m32+m30*m01*m22+m20*m31*m02-m30*m21*m02-m20*m01*m32-m31*m22*m00);

				r.m20=m01*m12*m33+m31*m02*m13+m11*m32*m03-m31*m12*m03-m11*m02*m33-m32*m13*m01;
				r.m21=-(m00*m12*m33+m30*m02*m13+m10*m32*m03-m30*m12*m03-m10*m02*m33-m32*m13*m00);
				r.m22=m00*m11*m33+m30*m01*m13+m10*m31*m03-m30*m11*m03-m10*m01*m33-m31*m13*m00;
				r.m23=-(m00*m11*m32+m30*m01*m12+m10*m31*m02-m30*m11*m02-m10*m01*m32-m31*m12*m00);

				r.m30=-(m01*m12*m23+m21*m02*m13+m11*m22*m03-m21*m12*m03-m11*m02*m23-m22*m13*m01);
				r.m31=m00*m12*m23+m20*m02*m13+m10*m22*m03-m20*m12*m03-m10*m02*m23-m22*m13*m00;
				r.m32=-(m00*m11*m23+m20*m01*m13+m10*m21*m03-m20*m11*m03-m10*m01*m23-m21*m13*m00);
				r.m33=m00*m11*m22+m20*m01*m12+m10*m21*m02-m20*m11*m02-m10*m01*m22-m21*m12*m00;

		return r;
	 }



};
/////////////////+-/////////////////
template <class T>
inline mat4x4<T> operator+(const mat4x4<T>& mat1, const mat4x4<T>& mat2)
{
	mat4x4<T> result(mat1);
	result+= mat2;
	return result;
};

template <class T>
inline mat4x4<T> operator-(const mat4x4<T>& mat1, const mat4x4<T>& mat2)
{
	mat4x4<T> result(mat1);
	result-= mat2;
	return result;
};
/////////////////*^/////////////////
template <class T>
inline mat4x4<T> operator*(const mat4x4<T>& mat1, const mat4x4<T>& mat2)
{
	mat4x4<T> r;
	r.m00=mat1.m00*mat2.m00+mat1.m01*mat2.m10+mat1.m02*mat2.m20+mat1.m03*mat2.m30;
	r.m01=mat1.m00*mat2.m01+mat1.m01*mat2.m11+mat1.m02*mat2.m21+mat1.m03*mat2.m31;
	r.m02=mat1.m00*mat2.m02+mat1.m01*mat2.m12+mat1.m02*mat2.m22+mat1.m03*mat2.m32;
    r.m03=mat1.m00*mat2.m03+mat1.m01*mat2.m13+mat1.m02*mat2.m23+mat1.m03*mat2.m33;

	r.m10=mat1.m10*mat2.m00+mat1.m11*mat2.m10+mat1.m12*mat2.m20+mat1.m13*mat2.m30;
	r.m11=mat1.m10*mat2.m01+mat1.m11*mat2.m11+mat1.m12*mat2.m21+mat1.m13*mat2.m31;
	r.m12=mat1.m10*mat2.m02+mat1.m11*mat2.m12+mat1.m12*mat2.m22+mat1.m13*mat2.m32;
    r.m13=mat1.m10*mat2.m03+mat1.m11*mat2.m13+mat1.m12*mat2.m23+mat1.m13*mat2.m33;

	r.m20=mat1.m20*mat2.m00+mat1.m21*mat2.m10+mat1.m22*mat2.m20+mat1.m23*mat2.m30;
	r.m21=mat1.m20*mat2.m01+mat1.m21*mat2.m11+mat1.m22*mat2.m21+mat1.m23*mat2.m31;
	r.m22=mat1.m20*mat2.m02+mat1.m21*mat2.m12+mat1.m22*mat2.m22+mat1.m23*mat2.m32;
    r.m23=mat1.m20*mat2.m03+mat1.m21*mat2.m13+mat1.m22*mat2.m23+mat1.m23*mat2.m33;

	r.m30=mat1.m30*mat2.m00+mat1.m31*mat2.m10+mat1.m32*mat2.m20+mat1.m33*mat2.m30;
	r.m31=mat1.m30*mat2.m01+mat1.m31*mat2.m11+mat1.m32*mat2.m21+mat1.m33*mat2.m31;
	r.m32=mat1.m30*mat2.m02+mat1.m31*mat2.m12+mat1.m32*mat2.m22+mat1.m33*mat2.m32;
    r.m33=mat1.m30*mat2.m03+mat1.m31*mat2.m13+mat1.m32*mat2.m23+mat1.m33*mat2.m33;

	return r;
};
template <class T>
inline mat4x4<T> operator*(const mat4x4<T>& mat, double a)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}
template <class T>
inline mat4x4<T> operator*(const mat4x4<T>& mat, int a)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}


template <class T>
inline mat4x4<T> operator*(const mat4x4<T>& mat, float a)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}

template <class T>
inline mat4x4<T> operator*(const mat4x4<T>& mat, long a)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}
//-----------------
template <class T>
inline mat4x4<T> operator*(const long a, const mat4x4<T>& mat)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}
template <class T>
inline mat4x4<T> operator*(const int a, const mat4x4<T>& mat)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}

template <class T>
inline mat4x4<T> operator*(const double a, const mat4x4<T>& mat)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}
template <class T>
inline mat4x4<T> operator*(const float a, const mat4x4<T>& mat)
{
	mat4x4<T> r(mat);
	return r *= (T)a;
}

//*********************

template <class T>
inline mat4x4<T> operator/(const mat4x4<T>& mat, const long a)
{
	mat4x4<T> r(mat);
	return r/=(T)a;
}

template <class T>
inline mat4x4<T> operator/(const mat4x4<T>& mat, const double& a)
{
	mat4x4<T> r(mat);
	return r/=(T)a;
}
template <class T>
inline mat4x4<T> operator/(const mat4x4<T>& mat, const float& a)
{
	mat4x4<T> r(mat);
	return r/=(T)a;
}
///////////////////-


template <class T>
inline bool operator==(const mat4x4<T>& mat1, const mat4x4<T>& mat2)
{
	return ((mat1.m00 ==mat2.m00) && (mat1.m01 ==mat2.m01) && (mat1.m02 ==mat2.m02) && (mat1.m03 ==mat2.m03) &&
		   (mat1.m10 ==mat2.m10) && (mat1.m11 ==mat2.m11) && (mat1.m12 ==mat2.m12) && (mat1.m13 ==mat2.m13) &&
		   (mat1.m20 ==mat2.m20) && (mat1.m21 ==mat2.m21) && (mat1.m22 ==mat2.m22) && (mat1.m23 ==mat2.m23) &&
		   (mat1.m30 ==mat2.m30) && (mat1.m31 ==mat2.m31) && (mat1.m32 ==mat2.m32) && (mat1.m33 ==mat2.m33));
};

template <class T>
inline bool operator!=(const mat4x4<T>& v1, const mat4x4<T>& v2)
{
	return !operator==(v1,v2);
};


typedef mat4x4<double> d4x4;
typedef mat4x4<float> f4x4;

#endif
